/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

typedef enum	ptype_s		ptype_t;
typedef struct	particle_s	particle_t;

enum ptype_s
{
	pt_static, 
	pt_fire, 
	pt_bubble, 
	pt_explode, 
	pt_blood,
	pt_blood2,
	pt_snow,
	pt_rain,
	pt_bulletpuff, 
	pt_fade,
	pt_grav,
	pt_rail,
	pt_blob,
	pt_blob2,
	pt_smokeexp
};

struct particle_s
{
	int					texnum;
	float				die;
	float				time;
	float				color;
	float				scale;
	float				alpha;
	float				bounce;
	vec3_t				origin;
	vec3_t				oldorigin;
	vec3_t				velocity;
	ptype_t				type;

	particle_t			*next;
	particle_t			*prev;
};

int		part_tex;
int		smoke1part_tex;
int		smoke2part_tex;
int		smoke3part_tex;
int		smoke4part_tex;
int		bloodpart_tex;
int		bubblepart_tex;
int		snowpart_tex;
int		rainpart_tex;

particle_t	*active_particles, *free_particles, *particles;

int		r_numparticles;

vec3_t	r_pright, r_pup, r_ppn;

extern byte	particle[32][32];
extern byte	smoke1[32][32];
extern byte	smoke2[32][32];
extern byte	smoke3[32][32];
extern byte	smoke4[32][32];
extern byte	blood[32][32];
extern byte	bubble[32][32];
extern byte	snow[32][32];
extern byte	rain[32][32];

void R_InitParticleTexture (void)
{
	int		x, y;
	byte	data[32][32][4];

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= particle[x][y];
		}
	}
	part_tex = GL_LoadTexture ("particle", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= smoke1[x][y];
		}
	}
	smoke1part_tex = GL_LoadTexture ("smoke1", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= smoke2[x][y];
		}
	}
	smoke2part_tex = GL_LoadTexture ("smoke2", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= smoke3[x][y];
		}
	}
	smoke3part_tex = GL_LoadTexture ("smoke3", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= smoke4[x][y];
		}
	}
	smoke4part_tex = GL_LoadTexture ("smoke4", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= blood[x][y];
		}
	}
	bloodpart_tex = GL_LoadTexture ("blood", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= bubble[x][y];
		}
	}
	bubblepart_tex = GL_LoadTexture ("bubble", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= snow[x][y];
		}
	}
	snowpart_tex = GL_LoadTexture ("snow", 32, 32, &data[0][0][0], true, true, 4);

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= rain[x][y];
		}
	}
	rainpart_tex = GL_LoadTexture ("rain", 32, 32, &data[0][0][0], true, true, 4);
}

/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int		i = COM_CheckParm ("-particles");

	r_numparticles = 4096;

	if (i)
	{
		r_numparticles = (int)(atoi(com_argv[i+1]));
	}

	particles = (particle_t *) Hunk_AllocName (r_numparticles * sizeof(particle_t), "particles");

	memset(particles, 0, r_numparticles * sizeof(particle_t));

	for( i = 0 ; i < r_numparticles ; ++i )
	{
		particles[i].prev	= &particles[i - 1];
		particles[i].next	= &particles[i + 1];
	}

	particles[0].prev		= NULL;
	particles[r_numparticles - 1].next	= NULL;

	free_particles			= &particles[0];
	active_particles		= NULL;

	R_InitParticleTexture ();
}

/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;

	free_particles			= &particles[0];
	active_particles		= NULL;

	memset(particles, 0, r_numparticles * sizeof(particle_t));

	for( i = 0 ; i < r_numparticles ; ++i )
	{
		particles[i].prev	= &particles[i - 1];
		particles[i].next	= &particles[i + 1];
	}

	particles[0].prev		= NULL;
	particles[r_numparticles - 1].next	= NULL;
}

particle_t* addParticle()
{
	particle_t*	p	= free_particles;

	if(!p)
	{
		return NULL;
	}

	free_particles	= p->next;

	if(free_particles)
	{
		free_particles->prev	= NULL;
	}

	p->next			= active_particles;

	if(active_particles)
	{
		active_particles->prev	= p;
	}

	active_particles			= p;

	return p;
}

particle_t* remParticle(particle_t*	p)
{
	particle_t*	next;

	if(!p)
	{
		return NULL;
	}

	if(p == active_particles)
	{
		active_particles	= p->next;
	}

	if(p->next)
	{
		p->next->prev		= p->prev;
	}

	if(p->prev)
	{
		p->prev->next		= p->next;
	}

	if(free_particles)
	{
		free_particles->prev	= p;
	}

	next					= p->next;

	p->next					= free_particles;

	free_particles			= p;

	p->prev	= NULL;

	return next;
}

void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	origin;
	int		r;
	int		c;
	particle_t	*p;
	char	name[MAX_OSPATH];

	sprintf (name,"maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);
	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}
	
	Con_Printf ("Reading %s...\n", name);
	c = 0;
	for ( ;; )
	{
		r = fscanf (f,"%f %f %f\n", &origin[0], &origin[1], &origin[2]);
		if (r != 3)
			break;
		c++;
		
		p	= addParticle();
		if(!p)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}
		
		p->texnum			= part_tex;
		p->bounce			= 0;
		p->scale			= 2;
		p->alpha			= 200;
		p->die				= 99;
		p->color			= (-c)&15;
		p->type				= pt_static;

		VectorCopy (vec3_origin, p->velocity);
		VectorCopy (origin, p->origin);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

/*
===============
R_EntityParticles
===============
*/

float	r_avertexnormals[162][3] = {
#include "anorms.h"
};

vec3_t			avelocities[128];

void R_EntityParticles (entity_t *ent)
{
	int			i;
	float		sp, sy, cp, cy, angle;
	vec3_t		forward;
	particle_t	*p;

	if (!avelocities[0][0])
	{
		for (i=0 ; i<384 ; i++)
		{
			avelocities[0][i] = (rand() & 255) * 0.01;
		}
	}

	for (i=0 ; i<128 ; i++)
	{
		angle				= cl.time * avelocities[i][0];
		sy					= sin(angle);
		cy					= cos(angle);

		angle				= cl.time * avelocities[i][1];
		sp					= sin(angle);
		cp					= cos(angle);
	
		forward[0]			= cp*cy;
		forward[1]			= cp*sy;
		forward[2]			= -sp;

		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->scale			= 2;
		p->alpha			= 200;
		p->die				= cl.time + 0.01;

		p->texnum			= part_tex;
		p->bounce			= 0;
		p->color			= 111;
		p->type				= pt_static;
		
		p->origin[0]		= ent->origin[0] + r_avertexnormals[i][0]*64 + forward[0]*16;			
		p->origin[1]		= ent->origin[1] + r_avertexnormals[i][1]*64 + forward[1]*16;			
		p->origin[2]		= ent->origin[2] + r_avertexnormals[i][2]*64 + forward[2]*16;			
	}
}

/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		origin, direction;
	int			count, color;

	origin[0]		= MSG_ReadCoord ();
	origin[1]		= MSG_ReadCoord ();
	origin[2]		= MSG_ReadCoord ();

	direction[0]	= MSG_ReadChar () * 0.0625;
	direction[1]	= MSG_ReadChar () * 0.0625;
	direction[2]	= MSG_ReadChar () * 0.0625;

	count			= MSG_ReadByte ();
	color			= MSG_ReadByte ();

	R_RunParticleEffect (origin, direction, color, count);
}
	
/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t origin)
{
	int			i, contents;
	particle_t	*p;

	for (i=0 ; i<64 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->scale			= 2;

		p->alpha			= 200;
		p->texnum			= smoke1part_tex + (rand() & 3);
		p->bounce			= 0;
		p->type				= pt_smokeexp;

		p->velocity[0]		= (rand() & 3) - 2;
		p->velocity[1]		= (rand() & 3) - 2;
		p->velocity[2]		= (rand() & 3) - 2;

		p->color			= (rand() & 7) + 105;
		p->die				= cl.time + 5;

		p->origin[0]		= origin[0] + (rand() &31) - 16;
		p->origin[1]		= origin[1] + (rand() &31) - 16;
		p->origin[2]		= origin[2] + (rand() &31) - 16;
	}

	for (i=0 ; i<256 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		contents			= Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;
		p->die				= cl.time + 5;

		if ((contents		== CONTENTS_EMPTY) ||
			(contents		== CONTENTS_SOLID))
		{
			p->texnum		= part_tex;
			p->bounce		= 1.5;
			p->color		= 111;
			p->type			= pt_explode;
		}
		else
		{
			p->texnum		= bubblepart_tex;
			p->bounce		= 0;
			p->color		= 244;
			p->type			= pt_bubble;
		}

		p->origin[0]		= origin[0] + ((rand() & 31) - 16);
		p->origin[1]		= origin[1] + ((rand() & 31) - 16);
		p->origin[2]		= origin[2] + ((rand() & 31) - 16);

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;
	}
}

/*
===============
R_ParticleExplosion2

===============
*/
void R_ParticleExplosion2 (vec3_t origin, int colorStart, int colorLength)
{
	int			i,contents;
	particle_t	*p;
	int			colorMod = 0;

	for (i=0; i<384; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		contents			= Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;
		p->die				= cl.time + 5;

		if ((contents		== CONTENTS_EMPTY) ||
			(contents		== CONTENTS_SOLID))
		{
			p->texnum		= part_tex;
			p->bounce		= 0;
			p->color		= colorStart + (colorMod % colorLength);
			colorMod++;
			p->type			= pt_explode;
		}
		else
		{
			p->texnum		= bubblepart_tex;
			p->bounce		= 0;
			p->color		= 244;
			p->type			= pt_bubble;
		}

		p->origin[0]		= origin[0] + ((rand() & 31) - 16);
		p->origin[1]		= origin[1] + ((rand() & 31) - 16);
		p->origin[2]		= origin[2] + ((rand() & 31) - 16);

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;
	}
}

/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosion (vec3_t origin)
{
	int			i;
	particle_t	*p;
	
	for (i=0 ; i<384 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->die				= cl.time + 1;

		p->type				= pt_blob;
		p->color			= (rand() & 7) + 66;

		p->origin[0]		= origin[0] + (rand() & 31) - 16;
		p->origin[1]		= origin[1] + (rand() & 31) - 16;
		p->origin[2]		= origin[2] + (rand() & 31) - 16;

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;
		
		if (i & 1)
		{
			p->type			= pt_blob2;
			p->color		= (rand()  & 7) + 150;
		}
	}
}

/*
===============
R_RunParticleEffect

===============
*/

void R_RunParticleEffect (vec3_t origin, vec3_t direction, int color, int count)
{
	int			i, contents;
	particle_t	*p;

	if (count == 128)
	{
		for (i=0 ; i<count ; i++)
		{	// rocket explosion
			p	= addParticle();
			if(!p)
			{
				return;
			}

			contents			= Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

			p->scale			= 2;
			p->alpha			= 200;
			p->die				= cl.time + 5;
			
			if ((contents		== CONTENTS_EMPTY) ||
				(contents		== CONTENTS_SOLID))
			{
				p->texnum		= part_tex;
				p->bounce		= 0;
				p->color		= 252;
				p->type			= pt_explode;
			}
			else
			{
				p->texnum		= bubblepart_tex;
				p->bounce		= 0;
				p->color		= 244;
				p->type			= pt_bubble;
			}

			p->origin[0]		= origin[0] + ((rand() & 31) - 16);
			p->origin[1]		= origin[1] + ((rand() & 31) - 16);
			p->origin[2]		= origin[2] + ((rand() & 31) - 16);

			p->velocity[0]		= (rand() & 511) - 256;
			p->velocity[1]		= (rand() & 511) - 256;
			p->velocity[2]		= (rand() & 511) - 256;
		}
		return;
	}

	for (i=0 ; i<count ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->texnum			= bloodpart_tex;
		p->bounce			= 0;
		p->scale			= 2;
		p->alpha			= 200;
		p->die				= cl.time + 5;
		p->color			= (color&~7) + (rand()&7);
		p->type				= pt_blood2;

		p->origin[0]		= origin[0] + ((rand() & 15) - 8);
		p->origin[1]		= origin[1] + ((rand() & 15) - 8);
		p->origin[2]		= origin[2] + ((rand() & 15) - 8);

		p->velocity[0]		= direction[0] * 15;
		p->velocity[1]		= direction[1] * 15;
		p->velocity[2]		= direction[2] * 15;
	}
}

/*
===============
R_SparkShower

===============
*/
void R_SparkShower (vec3_t origin, vec3_t direction)
{
	int			i, contents;
	particle_t	*p;

	p	= addParticle();
	if (!free_particles)
	{
		return;
	}

	contents			= Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

	if ((contents		== CONTENTS_EMPTY) ||
		(contents		== CONTENTS_SOLID))
	{
		p->scale			= 2;
		p->alpha			= 200;
		p->texnum			= smoke1part_tex + (rand() & 3);
		p->bounce			= 0;
		p->type				= pt_bulletpuff;
		p->color			= 12;
		p->die				= cl.time + 5;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->velocity[0]		= (rand() & 7) - 4;
		p->velocity[1]		= (rand() & 7) - 4;
		p->velocity[2]		= (rand() & 7) - 4;
	}

	for (i=0 ; i<10 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		contents			= Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

		p->scale			= (rand() & 3) +1;

		p->alpha			= 200;
		p->die				= cl.time + 5;

		if ((contents		== CONTENTS_EMPTY) ||
			(contents		== CONTENTS_SOLID))
		{
			p->texnum		= part_tex;
			p->bounce		= 1.5;
			p->color		= 252;
			p->type			= pt_explode;
		}
		else
		{
			p->texnum		= bubblepart_tex;
			p->bounce		= 0;
			p->color		= 244;
			p->type			= pt_bubble;
		}

		p->origin[0]		= origin[0] + ((rand() & 7) - 4);
		p->origin[1]		= origin[1] + ((rand() & 7) - 4);
		p->origin[2]		= origin[2] + ((rand() & 7) - 4);

		p->velocity[0]		= direction[0] + (rand() & 127) - 64;
		p->velocity[1]		= direction[1] + (rand() & 127) - 64;
		p->velocity[2]		= direction[2] + (rand() & 127) - 64;
	}
}

/*
===============
R_Snow
===============
*/
void R_Snow (vec3_t min, vec3_t max, int flakes)
{
	int			i;
	vec3_t		difference;
	particle_t	*p;

	for (i=0 ; i<flakes ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		VectorSubtract(max, min, difference);

		p->die				= cl.time + 10;

		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;
		p->texnum			= snowpart_tex;
		p->bounce			= 0;
		p->type				= pt_snow;
		p->color			= 15;

		p->origin[0]		= difference[0] * (rand () & 2047) * 0.00048828125 + min[0];	// Tomaz - Speed
		p->origin[1]		= difference[1] * (rand () & 2047) * 0.00048828125 + min[1];	// Tomaz - Speed
		p->origin[2]		= max[2] - 10;

		p->velocity[0]		= 0;
		p->velocity[1]		= 0;
		p->velocity[2]		= -50;
	}
}

/*
===============
R_Rain
===============
*/
void R_Rain (vec3_t min, vec3_t max, int drops)
{
	int			i;
	vec3_t		difference;
	particle_t	*p;

	for (i=0 ; i<drops ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		VectorSubtract(max, min, difference);

		p->die				= cl.time + 10;

		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;

		p->texnum			= rainpart_tex;
		p->bounce			= 0;
		p->type				= pt_rain;
		p->color			= 15;

		p->origin[0]		= difference[0] * (rand () & 2047) * 0.00048828125 + min[0];	// Tomaz - Speed
		p->origin[1]		= difference[1] * (rand () & 2047) * 0.00048828125 + min[1];	// Tomaz - Speed
		p->origin[2]		= max[2] - 10;

		p->velocity[0]		= 0;
		p->velocity[1]		= 0;
		p->velocity[2]		= -400;
	}
}

/*
===============
R_LavaSplash
===============
*/
void R_LavaSplash (vec3_t origin)
{
	int			i, j;
	particle_t	*p;

	for (i=-16 ; i<16 ; i++)
	{	
		for (j=-16 ; j<16 ; j++)
		{
			p	= addParticle();
			if(!p)
			{
				return;
			}
		
			p->texnum			= part_tex;
			p->bounce			= 0;
			p->scale			= 2;
			p->alpha			= 200;
			p->die				= cl.time + 5;
			p->color			= 231;
			p->type				= pt_grav;
				
			p->velocity[0]		= (rand() & 63 + 50) * (j*8 + (rand()&7));
			p->velocity[1]		= (rand() & 63 + 50) * (i*8 + (rand()&7));
			p->velocity[2]		= (rand() & 63 + 50) * 256;
	
			p->origin[0]		= origin[0] + (rand() & 7) + (j * 8);
			p->origin[1]		= origin[1] + (rand() & 7) + (i * 8);
			p->origin[2]		= origin[2] + (rand() & 63);

		}
	}
}

/*
===============
R_TeleportSplash

===============
*/
void R_TeleportSplash (vec3_t origin)
{
	int			i, j, k;
	particle_t	*p;

	for (i=-16 ; i<16 ; i+=4)
	{	
		for (j=-16 ; j<16 ; j+=4)
		{
			for (k=-24 ; k<32 ; k+=4)
			{
				p	= addParticle();
				if(!p)
				{
					return;
				}
		
				p->texnum			= part_tex;
				p->bounce			= 0;
				p->scale			= 2;
				p->alpha			= 200;
				p->die				= cl.time + 1;
				p->color			= 254;
				p->type				= pt_fade;
	
				p->origin[0]		= origin[0] + i + (rand() & 7);
				p->origin[1]		= origin[1] + j + (rand() & 7);
				p->origin[2]		= origin[2] + k + (rand() & 7);
	
				p->velocity[0]		= i*2 + (rand() & 31) - 16;
				p->velocity[1]		= j*2 + (rand() & 31) - 16;
				p->velocity[2]		= k*2 + (rand() & 31) + 24;
			}
		}
	}
}

void R_RailTrail (vec3_t start, vec3_t end, vec3_t angle)
{
	vec3_t		vec;
	float		len;
	vec3_t		forward, right, up;
	particle_t	*p;
	qboolean	rail = false;

	VectorSubtract (end, start, vec);

	len = VectorNormalize(vec);

	while (len > 0)
	{
		{  
			p	= addParticle();
			if(!p)
			{
				return;
			}

			p->alpha			= 200;
			p->scale			= 2;
			p->die				= cl.time + 1;
			p->color			= 15;
			p->type				= pt_rail;
			p->texnum			= part_tex;
			p->bounce			= 0;

			p->origin[0]		= start[0];
			p->origin[1]		= start[1];
			p->origin[2]		= start[2];

			p->velocity[0]		= 0;
			p->velocity[1]		= 0;
			p->velocity[2]		= 0;
		}

		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->alpha			= 200;
		p->scale			= 4;
		p->die				= cl.time + 1;
		p->color			= (rand() & 7) + 208;
		p->type				= pt_rail;
		p->texnum			= part_tex;
		p->bounce			= 0;

		AngleVectors (angle, forward, right, up);

		p->origin[0]		= start[0] + right[0] * 3;
		p->origin[1]		= start[1] + right[1] * 3;
		p->origin[2]		= start[2] + right[2] * 3;
		
		p->velocity[0]		= right[0] * 10;
		p->velocity[1]		= right[1] * 10;
		p->velocity[2]		= right[2] * 10;

		angle[2] += 5;

		len--;
		VectorAdd (start, vec, start);
	}                                     
}

void R_RocketTrail (vec3_t start, vec3_t end, entity_t *ent)
{
	int			contents;
	particle_t	*p;

	if (ent->debris_smoke <= 0)
		return;

	contents = Mod_PointInLeaf(start, cl.worldmodel)->contents;

	if (contents == CONTENTS_SKY || contents == CONTENTS_LAVA)
		return;

	if (cl.time > ent->time_left)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		VectorCopy(start, p->origin);

		if (contents == CONTENTS_EMPTY || contents == CONTENTS_SOLID)
		{
			p->die			= cl.time + 10;
			p->color		= 12;
			p->alpha		= ent->debris_smoke * 28;
			p->scale		= 2;
			p->texnum		= smoke1part_tex + (rand() & 3);
			p->type			= pt_fire;
			p->bounce			= 0;

			p->velocity[0]	= (rand() & 15) - 8;
			p->velocity[1]	= (rand() & 15) - 8;
			p->velocity[2]	= (rand() & 31) + 15;
			
			if (swdf.value)
			{
				p->velocity[0]	= (rand() & 7) - 4;
				p->velocity[1]	= (rand() & 7) - 4;
				p->velocity[2]	= (rand() & 7) - 4;
			}
		}
		else
		{
			p->die			= cl.time + 10;
			p->color		= 244;
			p->alpha		= 200;
			p->scale		= (rand() & 3) +1;
			p->texnum		= bubblepart_tex;
			p->bounce		= 0;
			p->type			= pt_bubble;

			p->velocity[0]	= 0;
			p->velocity[1]	= 0;
			p->velocity[2]	= 20;
		}
		ent->time_left		= cl.time + 0.02;
		ent->debris_smoke	-= 0.05f;
	}
}

void R_BloodTrail (vec3_t start, vec3_t end, entity_t *ent)
{
	particle_t	*p;

	if (cl.time > ent->time_left)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->die				= cl.time + 15;
		p->color			= (rand() & 3) + 68;
		p->alpha			= 200;
		p->scale			= 4;
		p->texnum			= bloodpart_tex;
		p->bounce			= 0;
		p->type				= pt_blood;

		p->velocity[0]		= (rand() & 15) - 8;
		p->velocity[1]		= (rand() & 15) - 8;
		p->velocity[2]		= (rand() & 15) - 8;

		p->origin[0]		= start[0] + ((rand() & 3) - 2);
		p->origin[1]		= start[1] + ((rand() & 3) - 2);
		p->origin[2]		= start[2] + ((rand() & 3) - 2);

		ent->time_left		= cl.time + 0.02;
	}
}

void R_TracerTrail (vec3_t start, vec3_t end, entity_t *ent, byte color)
{
	vec3_t		vec;
	static int	tracercount;
	particle_t	*p;
	
	VectorSubtract (end, start, vec);

	if (cl.time > ent->time_left)
	{
		{
			p	= addParticle();
			if(!p)
			{
				return;
			}

			VectorCopy (start, p->origin);

			p->die				= cl.time + 0.3;
			p->color			= color;
			p->alpha			= 200;
			p->scale			= 3;
			p->texnum			= part_tex;
			p->bounce			= 0;
			p->type				= pt_static;

			p->velocity[0]		= 5 * -vec[1];
			p->velocity[1]		= 5 *  vec[0];
			p->velocity[2]		= 0;
		}

		{
			p	= addParticle();
			if(!p)
			{
				return;
			}

			VectorCopy (start, p->origin);

			p->die				= cl.time + 0.3;
			p->color			= color;
			p->alpha			= 200;
			p->scale			= 3;
			p->texnum			= part_tex;
			p->bounce			= 0;
			p->type				= pt_static;

			p->velocity[0]		= 5 *  vec[1];
			p->velocity[1]		= 5 * -vec[0];
			p->velocity[2]		= 0;
		}
		ent->time_left			= cl.time + 0.02;
	}
}

void R_VoorTrail (vec3_t start, vec3_t end, entity_t *ent)
{
	particle_t	*p;

	if (cl.time > ent->time_left)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->die				= cl.time + 2;
		p->color			= 144;
		p->alpha			= 200;
		p->scale			= 3;
		p->texnum			= part_tex;
		p->bounce			= 0;
		p->type				= pt_fade;

		p->velocity[0]		= (rand() & 15) - 8;
		p->velocity[1]		= (rand() & 15) - 8;
		p->velocity[2]		= (rand() & 15) - 8;

		p->origin[0]		= start[0] + ((rand() & 3) - 2);
		p->origin[1]		= start[1] + ((rand() & 3) - 2);
		p->origin[2]		= start[2] + ((rand() & 3) - 2);

		ent->time_left		= cl.time + 0.02;
	}
}

float TraceLine (vec3_t start, vec3_t end, vec3_t impact, vec3_t normal);

/*
===============
R_DrawParticles
===============
*/

extern	cvar_t	sv_gravity;

void R_DrawParticles (void)
{
	particle_t	*p;

	int			texnum, contents;

	byte		*color24, alpha;

	float		frametime		= cl.time - cl.oldtime;
	float		time			= frametime * 10;
	float		fastgrav		= frametime * sv_gravity.value * 0.5;
	float		slowgrav		= fastgrav * 0.1;

	float		ft16			= frametime * 16;
	float		ft64			= frametime * 64;
	float		ft128			= frametime * 128;
	float		ft256			= frametime * 256;

	vec3_t		up				= { vup[0] * 0.75f, vup[1] * 0.75f, vup[2] * 0.75f };
	vec3_t		right			= { vright[0] * 0.75f, vright[1] * 0.75f, vright[2] * 0.75f };

	vec3_t		coord[4];

	vec3_t		v;

	if (!active_particles)
		return;

	texnum = part_tex;
    glBindTexture (GL_TEXTURE_2D, texnum);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDepthMask(false);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	glBegin (GL_QUADS);

	p = active_particles;

	while(p)
	{
		contents = Mod_PointInLeaf(p->origin, cl.worldmodel)->contents;

		if ((contents == CONTENTS_SOLID)||
			(contents == CONTENTS_SKY)	||
			(p->die < cl.time))
		{
			p = remParticle(p);
			continue;
		}

		if (p->texnum != texnum)
		{
			texnum = p->texnum;
			glEnd();
			glBindTexture (GL_TEXTURE_2D, texnum);
			glBegin(GL_QUADS);
		}

		VectorCopy(p->origin, p->oldorigin);

		p->origin[0] += p->velocity[0] * frametime;
		p->origin[1] += p->velocity[1] * frametime;
		p->origin[2] += p->velocity[2] * frametime;

		if (p->bounce)
		{
			vec3_t normal;
			float dist;
			if (TraceLine(p->oldorigin, p->origin, v, normal) < 1)
			{
				VectorCopy(v, p->origin);

				dist = DotProduct(p->velocity, normal) * -p->bounce;
				VectorMA(p->velocity, dist, normal, p->velocity);
				if (DotProduct(p->velocity, p->velocity) < 0.03)
				{
					VectorClear(p->velocity);
					p->bounce = 0;
				}
			}
		}

		switch (p->type)
		{
		case pt_static:
			break;

		case pt_fire:

			if(!swdf.value)
				p->scale += ft16;
			else
				p->scale += frametime * 12;

			p->alpha -= ft256;

			if(p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_bubble:

			p->velocity[0] = (rand() & 15) - 8;
			p->velocity[1] = (rand() & 15) - 8;
			p->velocity[2] = (rand() & 31) + 64;

			if (contents == CONTENTS_EMPTY)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_explode:
			p->velocity[2]	-= fastgrav;
			p->alpha		-= ft128;
			p->scale		-= frametime*2;

			if (contents != CONTENTS_EMPTY ||	p->alpha < 0 || p->scale < 0)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_blood:
			p->velocity[2] -= slowgrav;
			break;

		case pt_blood2:
			p->alpha -= ft128;
			p->velocity[2] -= fastgrav;
			if (p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_snow:
			if (cl.time > p->time)
			{
				p->time = cl.time + (rand() & 3) * 0.1;
				p->velocity[0] = (rand() & 31) - 16;
				p->velocity[1] = (rand() & 31) - 16;
			}
			break;			

		case pt_rain:
			break;

		case pt_bulletpuff:

			if(!swdf.value)
				p->scale += ft16;
			else
				p->scale += frametime * 12;

			p->alpha -= ft256;
			if (contents != CONTENTS_EMPTY || p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_smokeexp:
			p->scale += ft16;
			p->alpha -= ft256;
			if (contents != CONTENTS_EMPTY || p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_fade:
			p->alpha -= ft256;
			p->velocity[2] -= slowgrav;

			if (p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;

		case pt_grav:
			p->velocity[2] -= slowgrav;
			break;

		case pt_rail:
			p->alpha -= ft256;

			if (p->alpha < 1)
			{
				p = remParticle(p);
				continue;
			}
			break;
		}

		alpha	= p->alpha;

		color24 = (byte *) &d_8to24table[(int)p->color];

		glColor4ub(color24[0], color24[1], color24[2], alpha);

		__asm
		{
			//	move p into eax for pointer derefrencing
			mov				eax,			p

			//	load scale first - its the most used
			fld				dword ptr [eax]particle_t.scale

			//	load right[0..2] multiplying each by scale (st(0))
			fld				dword ptr right
			fmul			st(0),			st(1)						//	below, right[0] = st(6)

			fld				dword ptr right + 4
			fmul			st(0),			st(2)						//	right[1] = st(5)

			fld				dword ptr right + 8
			fmul			st(0),			st(3)						//	right[2] = st(4)

			//	load up[0..2] multiplying each by scale (st(3))
			fld				dword ptr up
			fmul			st(0),			st(4)						//	up[0] = st(3)

			fld				dword ptr up + 4
			fmul			st(0),			st(5)						//	up[1] = st(2)

			fld				dword ptr up + 8
			fmul			st(0),			st(6)						//	up[2] = st(1)

			//	we're done with scale, so free it
			ffree			st(6)

			//
			//	calculate coord[*][0]
			//

			//	first [0][0] and [3][0] w/ positive up factor
			fld				dword ptr [eax]particle_t.origin			//	load origin
			fadd			st(0),			st(3)						//	add in up[0]
			fld				st(0)										//	copy for coord[0][0]
			fadd			st(0),			st(7)						//	add right[0]
			fstp			dword ptr coord								//	store out to coord[0][0]
			fsub			st(0),			st(6)						//	sub right[0] for coord[3][0]
			fstp			dword ptr coord + 36						//	store out to coord[3][0]

			//	now [1][0] and [2][0] w/ negative up
			fld				dword ptr [eax]particle_t.origin			//	load origin
			fsub			st(0),			st(3)						//	sub up[0]
			fld				st(0)										//	copy for coord[1][0]
			fadd			st(0),			st(7)						//	add right[0]
			fstp			dword ptr coord + 12						//	store out to coord[1][0]
			fsub			st(0),			st(6)						//	sub right[0] for coord[2][0]
			fstp			dword ptr coord + 24						//	store out to coord[2][0]

			//
			//	calculate coord[*][1]
			//

			//	first [0][1] and [3][1] w/ positive up factor
			fld				dword ptr [eax]particle_t.origin + 4		//	load origin
			fadd			st(0),			st(2)						//	add in up[1]
			fld				st(0)										//	copy for coord[0][1]
			fadd			st(0),			st(6)						//	add right[1]
			fstp			dword ptr coord + 4							//	store out to coord[0][1]
			fsub			st(0),			st(5)						//	sub right[1] for coord[3][1]
			fstp			dword ptr coord + 40						//	store out to coord[3][1]

			//	now [1][1] and [2][1] w/ negative up
			fld				dword ptr [eax]particle_t.origin + 4		//	load origin
			fsub			st(0),			st(2)						//	sub up[1]
			fld				st(0)										//	copy for coord[1][1]
			fadd			st(0),			st(6)						//	add right[1]
			fstp			dword ptr coord + 16						//	store out to coord[1][1]
			fsub			st(0),			st(5)						//	sub right[1] for coord[2][1]
			fstp			dword ptr coord + 28						//	store out to coord[2][1]

			//
			//	calculate coord[*][2]
			//

			//	first [0][1] and [3][1] w/ positive up factor
			fld				dword ptr [eax]particle_t.origin + 8		//	load origin
			fadd			st(0),			st(1)						//	add in up[0]
			fld				st(0)										//	copy for coord[0][0]
			fadd			st(0),			st(5)						//	add right[0]
			fstp			dword ptr coord + 8							//	store out to coord[0][0]
			fsub			st(0),			st(4)						//	sub right[0] for coord[3][0]
			fstp			dword ptr coord + 44						//	store out to coord[3][0]

			//	now [1][1] and [2][1] w/ negative up
			fld				dword ptr [eax]particle_t.origin + 8		//	load origin
			fsub			st(0),			st(1)						//	sub up[2]
			fld				st(0)										//	copy for coord[1][2]
			fadd			st(0),			st(5)						//	add right[2]
			fstp			dword ptr coord + 20						//	store out to coord[1][2]
			fsub			st(0),			st(4)						//	sub right[2] for coord[2][2]
			fstp			dword ptr coord + 32						//	store out to coord[2][2]
			
			//
			//	free up up[] and right[]
			//

			ffree			st(0)										//	up[2]
			ffree			st(1)										//	up[1]
			ffree			st(2)										//	up[0]
			ffree			st(3)										//	right[2]
			ffree			st(4)										//	right[1]
			ffree			st(5)										//	right[0]

			//	done
		}

		glTexCoord2f (0,1);
		glVertex3fv (coord[0]);

		glTexCoord2f (0,0);
		glVertex3fv (coord[1]);

		glTexCoord2f (1,0); 
		glVertex3fv (coord[2]);

		glTexCoord2f (1,1);
		glVertex3fv (coord[3]);

		p = p->next;
	}

	glEnd ();
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDepthMask(true);
	glColor4f(1,1,1,1);
}